import sympy as sp
import unittest
from IntegralElimination import *
from IntegralElimination.reduction import reduction_M_by_P_simple_case

class TestIntegralElimination(unittest.TestCase):
    def test_14(self):
        x = sp.Function('x')
        y = sp.Function('y')
        t = sp.Symbol("t")
        theta = sp.Symbol("theta") 

        IA = IntegralAlgebra(order=[x(t),y(t)],
                            parameters=[theta])
        
        #is lm(P) is a monomial (not integral)
        P = IntegralPolynomial(IM(x(t))-IM(y(t)))
        M = IM(x(t)**2, y(t),x(t)*y(t))
        R = reduction_M_by_P_simple_case(IA,M,P)[1].get_sympy_repr()
        expected = IM(x(t)*y(t), y(t), x(t)*y(t))
        verify = sp.simplify(expected - R) == 0 
        self.assertTrue(verify) 

        #other case test 1 : M can't be reduced by p
        P = IntegralPolynomial(IM(1,x(t)*y(t))-IM(y(t)))   
        M = IM(x(t)**2, y(t),x(t)*y(t))
        R = reduction_M_by_P_simple_case(IA,M,P)[1]
        # using the anti fusion, you find N = x^2int(y(t))
        # lm(P)*N = x^2 int (y int (xy)) + x^2 int (xy int (y))
        # the leader is x^2 int (y int (xy)), which allow the reduction of M
        expected = IntegralPolynomial(
            IM(x(t)**2*y(t),y(t)) - IM(x(t)**2,x(t)*y(t),y(t))
        ) 
        verify = IA.polynomials_subtraction(R,expected).is_zero() 
        self.assertTrue(verify)

        #other case test 2
        P = IntegralPolynomial(IM(1,x(t)*y(t))-IM(y(t)))   
        M = IM(x(t)**2,x(t)*y(t))
        R = reduction_M_by_P_simple_case(IA,M,P)[1].get_sympy_repr()
        expected = IM(x(t)**2*y(t))
        verify = sp.simplify(expected - R) == 0 
        self.assertTrue(verify) 

        #last test
        P = IntegralPolynomial(IM(1,x(t)*y(t))-IM(y(t)))   
        M = IM(1,x(t)*y(t))
        R = reduction_M_by_P_simple_case(IA,M,P)[1].get_sympy_repr()
        expected = IM(y(t))
        verify = sp.simplify(expected - R) == 0 
        self.assertTrue(verify) 

if __name__ == '__main__':
    unittest.main()


import sympy as sp 

from .IntegralMonomial import IM

def LeadingMonomial(m1, m2, order):
    m1_dict =  m1.as_powers_dict()
    m1_tuple = tuple(m1_dict.get(var, 0) for var in order)
    m2_dict =  m2.as_powers_dict()
    m2_tuple = tuple(m2_dict.get(var, 0) for var in order)
    if m1_tuple > m2_tuple:
        return m1
    else:
        return m2 


# compare two monimials w.r.t to the plex order on X
# return true if w1<=w2

def comp_lexico(w1, w2, order):   
    """
    w1 or w2 can be sp.Expr or int
    """
    is_float_w1 = not isinstance(w1,sp.Expr) 
    is_float_w2 = not isinstance(w2,sp.Expr) 

    if (w1 == 0) : return True 
    elif (w2 == 0) : return False 
    elif is_float_w1 and is_float_w2 :
        return w1 <= w2
    elif is_float_w1 and not is_float_w2 :
        return True
    elif not is_float_w1 and is_float_w2 :
        return False
    
    ld = LeadingMonomial(w1, w2, order) 
    if  ld == w2 : 
        return True
    else: 
        return False
    
def comp_lexico_lists(L1,L2,order, reversed=False):
    # return true if L1<=L2
    if len(L1)==0:
        return True
    if len(L2)==0:
        return False
    #add padding
    n1 = len(L1)
    n2 = len(L2)
    n_max = max([n1,n2])
    L1 = [*L1, *[0]*(n_max-n1)]
    L2 = [*L2, *[0]*(n_max-n2)]
    for i in range(len(L1)):
        if L1[i] != L2[i]:
            if reversed:
                return not comp_lexico(L1[i], L2[i], order)
            return comp_lexico(L1[i], L2[i], order)
    return True

#integral monomials ordering
def IMO(M1: IM, M2: IM, order: list):   
    """
    M1 and M2 are integral monomials
    order : list of monomials
    IMO is an elimination ordering

    return True if i1 <= i2 
    """ 
    L1_c = M1.get_content()
    L2_c = M2.get_content()
    n1 = len(L1_c)
    n2 = len(L2_c)
    m1 = sp.Mul(*[L1_c[i] for i in range(n1)])
    m2 = sp.Mul(*[L2_c[i] for i in range(n2)]) 
    
    L1 = [m1, n1, *reversed(L1_c)] 
    L2 = [m2, n2, *reversed(L2_c)] 
    
    return comp_lexico_lists(L1,L2,order)


def IMO_elim_integ(M1: IM, M2: IM, order: list):   
    L1_c = M1.get_content()
    L2_c = M2.get_content()
    n1 = len(L1_c)
    n2 = len(L2_c)
    L1 = [n1, *reversed(L1_c)] 
    L2 = [n2, *reversed(L2_c)] 
    return comp_lexico_lists(L1,L2,order)


def IMO_elim_cst(M1: IM, M2: IM, order: list):   
    L1_c = M1.get_content()
    L2_c = M2.get_content()
    n1 = len(L1_c)
    n2 = len(L2_c) 
    m1 = sp.Mul(*[L1_c[i] for i in range(n1)])
    m2 = sp.Mul(*[L2_c[i] for i in range(n2)]) 
    L1 = [m1, n1, *reversed(L1_c)] 
    L2 = [m2, n2, *reversed(L2_c)] 
    return comp_lexico_lists(L1,L2,order,reversed=True)



import sympy as sp 

def is_float(expr):
    try:
        float(expr)
        return True
    except:
        return False

def is_int(expr):
    try:
        f = float(expr) 
        return f.is_integer()
    except:
        return False

def has_add_in_list(l):  
    for e in l:
        if not is_float(e) or not is_int(e):
            if e.has(sp.Add):
                return True
    return False
 

def expr_has_symbol(expr, symbol): 
    try: 
        return expr.has(symbol)
    except:
        return False


def expr_has_symbols(expr, symbols): 
    has_symbols = False
    for symbol in symbols:
        has_symbols = has_symbols or expr_has_symbol(expr,symbol)
    return has_symbols
 

def dicts_subtraction(L:list[dict]):
    assert len(L) == 2 #to admit the same signature as dicts_addition
    P,Q = L[0],L[1]
    PminusQ = P
    for N,c_N in Q.items(): 
        if N not in PminusQ:
            PminusQ[N] = -c_N
        else:
            PminusQ[N] -= c_N 
    keys_to_remove = [M for M, coeff in PminusQ.items() if coeff == 0]
    for key in keys_to_remove:
        del PminusQ[key] 
    return PminusQ

def dicts_addition(L:list[dict]):
    assert len(L) == 2 #to admit the same signature as dicts_addition
    P,Q = L[0],L[1]
    PpluqQ = P
    for N,c_N in Q.items(): 
        if N not in PpluqQ:
            PpluqQ[N] = c_N
        else:
            PpluqQ[N] += c_N 
    keys_to_remove = [M for M, coeff in PpluqQ.items() if coeff == 0]
    for key in keys_to_remove:
        del PpluqQ[key] 
    return PpluqQ

def dict_mul_by_coeff(d:dict, coeff:sp.Expr):
    for key in d:
        d[key] = d[key]*coeff
    return d


def subtract_lists(L1,L2): 
    if len(L2) > len(L1) or len(L1) == 0: return []
    if len(L2) == 0: return L1 
    res = []
    for i in range(len(L1)):
        if L2[0] == L1[i]:
            return res + subtract_lists(L1[:i]+L1[i+1:], L2[1:])
    return res     

def shuffle_list(l1, l2):
    """
    shuffle two lists
    u1.u >< v1.v = u1.(u >< v1.v) + v1.(u1.u >< v)
    with >< the shuffle operation
     
    return [u1, u2, ..., un] such that l1 >< l2 = u1 + u2 + ... + un,
    """
    res = []
    if len(l1) == 0:
        res = [l2]
    elif len(l2) == 0 :
        res = [l1]
    else:
        sh1 = shuffle_list(l1[1:], l2) # (u >< v1.v)
        sh2 = shuffle_list(l1, l2[1:]) # (u1.u >< v)
        for l in sh1:
            res = [*res, [l1[0], *l]] # u1.(u >< v1.v)
                
        for l in sh2:
            res = [*res, [l2[0], *l]] # v1.(u1.u >< v)
    return res


def lyndon_decomp(w:list, cmp_fn:callable, order:list):
    """
    w      is a word 
    cmp_fn is the comparison function used to compare two elements
            of the alphanets.
            In this context, we will use comp_lexico from ordering.py
    order  is the order used for comp
    """
    n=len(w)
    lyn_idx = []
    k = -1
    while k < n-1:
        i = k+1
        j = k+2
        while j < n and cmp_fn(w[i],w[j],order) :
            if w[i] == w[j]:
                i = i+1
            else :
                i = k+1
            j = j+1
        while k < i:
            lyn_idx += [k+1]
            k = k+j-i
    lyn_idx += [n]
    lyn_words =  [tuple(w[lyn_idx[i]:lyn_idx[i+1]])
                  for i in range(len(lyn_idx)-1)]
    return lyn_words

 


import sympy as sp
import unittest
from IntegralElimination import *

class TestIntegralElimination(unittest.TestCase):
    def test_13(self):
        x = sp.Function('x')
        y = sp.Function('y')
        t = sp.Symbol("t")
        theta = sp.Symbol("theta") 

        IA = IntegralAlgebra(order=[x(t),y(t)],
                            parameters=[theta])
        
        P = IntegralPolynomial(IM(1,x(t))-IM(y(t)))
        red_pow_2 = IA.reduced_power(P,2).get_sympy_repr() 
        
        expected = IM(y(t)**2) - 2*IM(1, x(t)*y(t))
        verify = sp.simplify(expected - red_pow_2) == 0 
 
        self.assertTrue(verify)

if __name__ == '__main__':
    unittest.main()


import sympy as sp
import unittest
from IntegralElimination import *

class TestIntegralElimination(unittest.TestCase):
    def test_10(self):
        x = sp.Function('x')
        y = sp.Function('y')
        t = sp.Symbol("t")
        theta = sp.Symbol("theta")
        P = IntegralPolynomial(3*IM(x(t))+theta*IM(1,y(t)))
         

        IA = IntegralAlgebra(order=[x(t),y(t)],
                            parameters=[theta])
        
        PQ = IA.integrate_polynomial(P).get_sympy_repr()
        
        expected = 3*IM(1,x(t))+theta*IM(1,1,y(t))
        verify = sp.simplify(expected - PQ) == 0
        self.assertTrue(verify)

if __name__ == '__main__':
    unittest.main()


import sympy as sp
import sympy.printing as printing
 
from .IntegralMonomial import IM
from .utils import is_float  


def task_simplify_coeff(M_coeff, cache_expand_coeff):
    try:
        M,coeff = M_coeff 
        if coeff not in cache_expand_coeff:  
            cache_expand_coeff[coeff] = sp.cancel(coeff)   
        expanded_coeff = cache_expand_coeff[coeff]
        return (M,expanded_coeff) 
    except: 
        raise RuntimeError

def init_coeff_dict(P, cache_expand_coeff={}, simplify_coeff=True):
    """
    P is a dict, a tuple, zero or a sp.Expr
    """
    if type(P) == dict or type(P)==tuple:
        tmp = {} 
        P=dict(P) 
        if simplify_coeff:   
            # args = [(M_coeff, cache_expand_coeff) for M_coeff in P.items()]
            # with ThreadPoolExecutor() as executor:
            #     res = executor.map(task_simplify_coeff, *zip(*args))   

            # sequential
            res = [task_simplify_coeff(
                    M_coeff,cache_expand_coeff) for M_coeff in P.items()] 

            for M, simplified_coeff in res:
                if simplified_coeff != 0:
                    tmp[M] = simplified_coeff
            
        else:
            tmp = P
    elif P == 0:
        tmp = {}
    elif type(P)==IM:
        tmp = {P:1}
    elif isinstance(P, sp.Expr):
        assert P.has(IM) 
        tmp = dict(P.as_coefficients_dict(IM))
    else:
        raise SyntaxError
    return tuple(tmp.items())


class IntegralPolynomial(): 
    # the cache is used to have the same id for two 
    # IntegralPolynomial that have the same content
    _cache = {} 
    _cache_expand_coeff = {}

    def __new__(cls, P, simplify_coeff=True): 
        """
        the cache system allow the check if a polynomial is in a set
        since the id of two integral polynomials with the same content
        will be the same
        """
        tmp = init_coeff_dict(P, 
                IntegralPolynomial._cache_expand_coeff, 
                simplify_coeff
        ) 
 
        # Check the cache to see if we already have this content
        if tmp in cls._cache:
            return cls._cache[tmp]

        # Create a new instance and cache it
        instance = super().__new__(cls)
        cls._cache[tmp] = instance
        return instance

    def __init__(self, 
                P,  
                simplify_coeff=True):  
        
        # Skip initialization if the instance is already initialized
        if hasattr(self, 'content'):
            return
        
        tmp = init_coeff_dict(P,
                IntegralPolynomial._cache_expand_coeff, 
                simplify_coeff
        )  
        self.content = tmp

    def get_content(self):
        """
        return an iterrable such that each item is a tuple (M,coeff)
        """
        return self.content

    def get_content_as_dict(self):
        return dict(self.get_content())


    def get_sympy_repr(self):
        """
        example : 
        a = IntegralPolynomial(IM(x(t))+theta*IM(1,x(t)))
        b = a.get_sympy(repr)
        print(b)
        print(type(b))
        >>> IM(x(t))+theta*IM(1,x(t))
        >>> type(sp.Expr)
        """
        L = self.get_content()
        sympy_repr = sp.Add(*[coeff*M for M,coeff in L]) 
        return sympy_repr
    
    def get_integral_repr(self):
        """ 
        example : 
        a = IntegralPolynomial(IM(x(t))+theta*IM(1,x(t)))
        b = a.get_sympy(repr)
        print(b)
        print(type(b))
        >>> x(t) + theta*int(x(t),t)
        >>> type(sp.Expr)
        """
        L = self.get_content()
        int_repr = sp.Add(*[sp.Mul(coeff,M.get_integral_repr()) for M,coeff in L]) 
        return int_repr
    
    def is_zero(self):
        return len(self.get_content()) ==0
    
    def __repr__(self):
        return f"IntegralPolynomial({self.get_sympy_repr()})"
    
    def repr_display_math(self):
        return '{}'.format(printing.latex(self.get_sympy_repr()))

    def _repr_latex_(self):  
        return '${}$'.format(printing.latex(self.get_sympy_repr()))
     
    def cut_P(self, cut_type: str):
        """
        Definition 8
        We simply extend the cut method of the Integral Monomial class to 
        integral polynomials 
        
        Disclaimer: this method will throw an error if you use it on 
        polynomial that can't be cutted 
        for exemple, trying to cut a pol with monomials of depth < 2 will 
        throw an error if you cut using i2+
        """
        P_cutted = {}
        L = self.get_content()
        for M,coeff in L:
            M_cutted = M.cut(cut_type) 
            if P_cutted.get(M_cutted) is None:
                P_cutted[M_cutted] = coeff 
            else:
                P_cutted[M_cutted] += coeff 
        return IntegralPolynomial(P_cutted)

    def get_P_I(self): 
        P_I = {}
        L = self.get_content()
        for M,coeff in L:
            M0 = M.cut("0") 
            if M.get_nb_int() >= 1 and M0 == IM(1):
                if M not in P_I:
                    P_I[M] = coeff
                else:
                    # we are not supposed to have this case
                    raise ValueError 
        # we avoid the simplification process of the coeffs
        # by using simplify_coeff=False
        return  IntegralPolynomial(P_I, simplify_coeff=False)

    def get_P_N(self): 
        P_N = {}
        L = self.get_content()
        for M,coeff in L: 
            M0 = M.cut("0")
            if M==IM(1): #cst
                P_N[M] = coeff
            elif M0 != IM(1):
                if M not in P_N:
                    P_N[M] = coeff
                else:
                    # we are not supposed to have this case
                    raise ValueError 
        # we avoid the simplification process of the coeffs
        # by using simplify_coeff=False
        return  IntegralPolynomial(P_N, simplify_coeff=False)
    
    def get_cst_terms(self):
        CST = {}
        L = self.get_content()
        for M,coeff in L:
            if M == IM(1):
                if M in CST:
                    CST[M] += coeff 
                else:
                    CST[M] = coeff 
        return  IntegralPolynomial(CST)
    
    def get_time_dependant_functions(self):
        """
        example: if you have 
        P = IntegralPolynomial(IM(x(t)) - x(0)*IM(1) 
                        - theta*IM(1,x(t)*y(t)**2) - IM(1,y(t)))
        it will return {x,y}
        """
        res = set()
        for f in self.get_sympy_repr().atoms(sp.Function):
            if f.func != IM and not is_float(f.args[0]):
                res.add(f) 
        return res


from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .IntegralAlgebra import IntegralAlgebra

import sympy as sp 
from ordered_set import OrderedSet

from .IntegralMonomial import IM
from .IntegralPolynomial import IntegralPolynomial 
from .utils import (
    is_int, 
)

from .check_conditions_LM import (
    check_condition_LM_half_reduced_product,
)


def reduction_M_by_P_simple_case(IA: IntegralAlgebra, 
                                 M: IM, 
                                 P: IntegralPolynomial,
                                 compute_W=True) -> tuple[list, IntegralPolynomial]:
    """
    Consider an integral monomial M
    and a nonzero integral polynomiaml P.
	
	If there exists an integral monomial 
    N such that lm{lm{P} · N}=M,
	then introduce
	R =  M-(1/l)P · N
	where l is the leading coefficient of P · N.

    This function returns a list W of "quotients"
    and the polynomial R 
    such that M = (sum of elements of the form (wi\int ... w_n P)) + R
    
    Here, the "sum of elements of the form (wi\int ... w_n P)" 
        is simply equal to (1/l)·N·P 
    Then, W = [ ((1/l)·N , P) ] where l is the leading coefficient of P·N
    """
    assert not P.is_zero() 

    LM_P,LC_P = IA.LM_LC(P)
    assert float(LC_P) == 1.0, "LC(P) should be normalized"

    N = IA.monomial_division(M,LM_P)
    if N == None:
        return None  
    
    PN = IA.polynomials_product(P,IntegralPolynomial(N))
    PN_norm,LC_PN = IA.normalize_LC_of_P(PN)
    
    assert M == IA.LM_LC(PN_norm)[0]

    R = IA.polynomials_subtraction(
            IntegralPolynomial(M),
            PN_norm
        ) 
    if not R.is_zero():
        LM_R = IA.LM_LC(R)[0]
        assert IA.compare_IM(LM_R,M)
    
    if not compute_W:
        W = []
        return W, R 
    W = [(IntegralPolynomial((1/LC_PN)*N),P),] 
    return W, R
 

def reduction_M_by_P_red_power_and_half_red_product(
            IA: IntegralAlgebra, 
            M: IM, 
            P: IntegralPolynomial,
            compute_W=True) -> tuple[list, IntegralPolynomial]:
    """
    Consider an integral monomial M and an integral polynomial P
    with P_I != 0, P_N != 0, and lm{P} = lm{P_I}$.
    
    Assume there exists a positive integer n
    and an integral monomial N such that
    - lm{P_I}[1] lm{lm{P_N}[0]}^{n-1} = M[1],
    - lm{ (P_I[2+]) · (P_N[1+])^{n-1} · N } = M[2+],
    - N[0] = 1.

    If the condition
    of check_condition_LM_half_reduced_product
    (this function checks the condition)
    holds for P^{\reduced_power{n}} half-reduced-prod N,  
    then introduce
	R = M - (1/l) M[0](P^{reduced_power{n}} half-reduced-prod N )
	where
	l is the leading coefficient 
    of M[0](P^{reduced_power{n}} half-reduced-prod N ).

    This function returns a list W of "quotients"
    and the polynomial R 
    such that M = (sum of elements of the form (wi\int ... w_n P)) + R

    Here, we want to write (1/l) M[0](P^{reduced_power{n}} half-reduced-prod N )
    as a sum of elements of the form (wi\int ... w_n P).

    1. P^{reduced_power{n}}                     = d0·P + int( d1·P )
        where d0 = sum_{k=1}^{n} (-1)^{k-1} binom(n,k) P_N ^{n-k} P^{k-1}
              d1 = sum_{k=1}^{n} n binom(n-1,k) P_I]1+] P_N ^{n-k-1} P^{k-1}

    2. P^{reduced_power{n}} half-reduced-prod N = d0·P·N + int( (d1·N-N]1+·d0)·P )
    
    3.  (1/l) M[0](P^{reduced_power{n}} half-reduced-prod N ) = 
            (1/l) M[0] d0·P·N +  (1/l) M[0] int( (d1·N-N]1+·d0)·P )

    We encode 3. as the list W = [ (1/l M0 d0 N, P), ((1/l)M0, d1 N - N]1+ d0, P) ] 
    """
    assert not P.is_zero() 
    LM_P,LC_P = IA.LM_LC(P)
    assert float(LC_P) == 1.0, "LC(P) should be normalized"
    if M.get_nb_int() == 0 : # |M| >= 1
        return None
    P_I = P.get_P_I()
    if P_I.is_zero(): return None 
    LM_P_I,LC_P_I = IA.LM_LC(P_I)
    if LM_P_I != LM_P : # we want lm(P) = lm(P_I)
        return None
    P_N = P.get_P_N()
    if P_N.is_zero(): return None 
    LM_P_N,LC_P_N = IA.LM_LC(P_N)
    M_1 = M.cut("1").get_content()[0]
    LM_P_N_0 = LM_P_N.cut("0").get_content()[0]
    LM_P_I_1 = LM_P_I.cut("1").get_content()[0]

    # we try to find n such that (lm(PI)[1]*lm(PN)[0]**(n-1))/M[1] = 1
    n = sp.Symbol("n_pow")
    pow_dict = ((LM_P_I_1*LM_P_N_0**(n-1))/M_1).as_powers_dict()
    if len(pow_dict) > 1: 
        #for example :
        # expr = ((x(t)*a(t))**n*y(t))/(x(t)**2*y(t))
        # expr.as_powers_dict()
        # {a(t)x(t):n, x(t): -2}
        return None 
    sp_solved = sp.solve(list(pow_dict.values())[0])
    if len(sp_solved)==0: return None
    solved_n = sp_solved[0] 
    if not(is_int(solved_n) and solved_n > 0):
        return None
    #we then have to verify the condition:
    #lm( (lm(P_I)[2+]) cdot (lm(P_N)[1+])**(n-1)) = M[2+]
    LM_P_I_i2plus = IntegralPolynomial(LM_P_I.cut("i2+"))
    LM_P_N_i1plus = IntegralPolynomial(LM_P_N.cut("i1+"))
    LM_P_N_i1plus_pow = IA.polynomial_power(LM_P_N_i1plus,solved_n-1)
    pol = IA.polynomials_product(LM_P_I_i2plus, LM_P_N_i1plus_pow)
    
    Mi2_plus = M.cut("i2+")
    B_P, c_B_P = IA.LM_LC(pol)
    N = IA.monomial_division(Mi2_plus, B_P)
    if N is None or N.get_content()[0] != 1:
        return None
    
    # we compute the reduced power
    P_reduced_pow_n = IA.reduced_power(P,solved_n)
    # we verify the condition on N
    if not check_condition_LM_half_reduced_product(IA, P_reduced_pow_n, N):
        return None
    # we can now compute R
    P_reduced_pow_n_red_prod_N = IA.half_reduced_product(P_reduced_pow_n, N)
    
    M0 = IntegralPolynomial(M.cut("0"))
    M0_P_pow_n = IA.polynomials_product(P_reduced_pow_n_red_prod_N, M0)
    M0_P_pow_n_norm, LC_M0_P_pow_n = IA.normalize_LC_of_P(M0_P_pow_n)
      
    assert M == IA.LM_LC(M0_P_pow_n_norm)[0]
    R = IA.polynomials_subtraction(IntegralPolynomial(M), 
                                  M0_P_pow_n_norm)
    if not R.is_zero():
        LM_R = IA.LM_LC(R)[0]
        assert IA.compare_IM(LM_R,M)

    if not compute_W:
        W = []
        return W, R 
    
    # we have M= 1/l M0 P^{circled(n)} half_red N + R
    # take H = 1/l M0 P^{circled(n)} halfred N 
    # here, H is equal to 1/l * M0( d0 N P + \int ((d1 N - N]1+ d0)P) )
    # with d0 and d1 defined as follows: 
    d0 = IntegralPolynomial(0)
    d1 = IntegralPolynomial(0) 
    for k in range(1,solved_n+1):  
        d0 = IA.polynomials_add(d0,  
                IA.product_P_Coeff( 
                    IA.polynomials_product(
                        IA.polynomial_power(P_N,solved_n-k),
                        IA.polynomial_power(P,k-1)
                    ),
                    (-1)**(k-1)*sp.binomial(solved_n,k)
                )
            )
                             
        d1 = IA.polynomials_add(d1,  
                IA.product_P_Coeff( 
                    IA.polynomials_product(
                        IA.polynomials_product(
                            IA.polynomial_power(P_N,solved_n-k-1),
                            IA.polynomial_power(P,k-1)
                        ),
                        P_I.cut_P("1+")
                    ),
                    solved_n*(-1)**(k-1)*sp.binomial(solved_n-1,k)
                )
            )
    # we want to return
    # W = [ (1/l M0 d0 N, P), ((1/l)M0, d1 N - N]1+ d0, P) ] 
    if N==IM(1): N1p = 0
    else : N1p = N.cut("1+")

    W0 = (IA.product_P_Coeff(
                    IA.polynomials_product(
                        IA.polynomials_product(M0, d0),
                                IntegralPolynomial(N)
                    ), 
                    1/LC_M0_P_pow_n),
            P,
        )

    W1 = (IA.product_P_Coeff(M0, 1/LC_M0_P_pow_n), 
            IA.polynomials_subtraction(
                IA.polynomials_product(d1,IntegralPolynomial(N)),
                IA.polynomials_product(d0,IntegralPolynomial(N1p))
            ),
            P,
        )
    W = [W0,W1] 
    return W, R 

def reduction_M_by_P(IA: IntegralAlgebra, 
                     M: IM, 
                     P: IntegralPolynomial,
                     compute_W=True
                    ) -> tuple[list, IntegralPolynomial]:
    """
    Lemma 14
    
    W is obtained by adding the prefix to the previous list 
        W = [ ((1/l)·N , P) ]
        or
        W = [ (1/l M0 d0 N, P), ((1/l)M0, d1 N - N]1+ d0, P) ] 
    computed by reduction_M_by_P_simple_case or reduction_M_by_P_red_power_and_half_red_product.
    See the functions add_prefix_to_quotients and mult_quotients_by_coeff.
    """
    assert not P.is_zero() 
    e = M.get_nb_int() 
    P_norm, LC_P = IA.normalize_LC_of_P(P)  
    for i in reversed(range(e+1)): 
        prefix = M.get_prefix(i)
        suffix = M.get_suffix(i)   
        res = reduction_M_by_P_simple_case(IA,
                    suffix, P_norm, compute_W=compute_W)  
        if res is not None: 
            W, R = res 
            R = IA.add_prefix_to_polynomial(prefix, R) 
            W = add_prefix_to_quotients(IA, prefix, W)   
            W = mult_quotients_by_coeff(IA, W, 1/LC_P)  
            #P has been multiplied by the previous line, we need to rewrite P without the coeffs
            W = [(*elem[:-1],P,)  for elem in W]  
            if not R.is_zero():
                LM_R = IA.LM_LC(R)[0]
                assert IA.compare_IM(LM_R,M)
            return W, R
        res = reduction_M_by_P_red_power_and_half_red_product(IA,
                    suffix, P_norm, compute_W=compute_W) 
        if res is not None:
            W, R = res
            R = IA.add_prefix_to_polynomial(prefix,R) 
            W = add_prefix_to_quotients(IA, prefix, W) 
            W = mult_quotients_by_coeff(IA, W, 1/LC_P) 
            W = [(*elem[:-1],P,) for elem in W]
            if not R.is_zero():
                LM_R = IA.LM_LC(R)[0]
                assert IA.compare_IM(LM_R,M)
            return W, R 
    return None

def reduce_LM(IA:IntegralAlgebra, 
           Q:IntegralPolynomial, 
           T:OrderedSet[IntegralPolynomial],
           compute_W:bool=True,
           W=None) -> tuple[list, IntegralPolynomial]:
    """
    reduce Q by the set of integral polynomials T
    until the LM cannot be reduced

      >>> i.e. this functions computes R
    and the list W of "quotients" such that 
    Q = sum(integral polynomials which uses elements of W) + R

    /!\ R is not the necessarly the normal form of Q w.r.t T
    """ 
    if W is None:
        W = []
    if Q.is_zero(): 
        return W, IntegralPolynomial(0)
    A = Q 

    LM_A,LC_A = IA.LM_LC(A)
    # test if LM can be reduced by a polynomial P of T
    if IA.storage["P_red_to_zero"].get(A):
        return IA.storage["P_red_to_zero"]
    for P in T:  
        res = reduction_M_by_P(IA, LM_A, P,compute_W)   
        if res is not None: 
            Wi, Ri = res 
            Wi = mult_quotients_by_coeff(IA,Wi,LC_A)

            # A=lc_A*lm(A)+B, i.e B is the tail, so we can just delete the leading 
            # monomial of A to obtain B=tail_A
            tail_A = A.get_content_as_dict()
            del tail_A[LM_A]

            # lm(A) has been reduced by P 
            # i.e. lm(A) = ... + Ri
            # and thus lc_A*lm(A) = lc_A*... + lc_A*Ri 
            LC_A_Ri = IA.product_P_Coeff(Ri, LC_A, return_dict=True)
            # A reduced by P:
            # A = ... + R, where R = tail_A+lc_A Ri
            R = IA.polynomials_add(tail_A, LC_A_Ri, return_dict=False)   
            W += Wi
            if R.is_zero():
                return W, R
            return reduce_LM(IA, R, T, compute_W=compute_W, W=W) 
    if A.is_zero():
        IA.storage["P_red_to_zero"][A] = IntegralPolynomial(0)
    return W, A

def reduce_full(IA:IntegralAlgebra, 
           Q:IntegralPolynomial, 
           T:OrderedSet[IntegralPolynomial],
           compute_W:bool=True,
           W=None) -> tuple[list, IntegralPolynomial]:
    """
    reduce Q by the set of integral polynomials T
    until until there are no more possible reductions

    >>> i.e. this functions computes the normal form R
    and the list W of "quotients" such that 
    Q = sum(integral polynomials which uses elements of W) + R
    """ 
    if W is None:
        W = []
    if Q.is_zero(): 
        return W, IntegralPolynomial(0)
    A = Q 
    if IA.storage["P_red_to_zero"].get(A):
        W, pol = IA.storage["P_red_to_zero"][A]
        return W, pol, True
    for Mi, alpha in A.get_content(): 
        for P in T:  
            res = reduction_M_by_P(IA,Mi,P,compute_W=compute_W) 
            if res is not None:
                Wi, Ri = res 
                Wi = mult_quotients_by_coeff(IA,Wi,alpha)
                Ri = IA.product_P_Coeff(Ri,alpha)
                tail_A = A.get_content_as_dict()
                del tail_A[Mi]
                tail_A = IntegralPolynomial(tail_A)
                R = IA.polynomials_add(tail_A, Ri)
                W += Wi 
                return reduce_full(IA, 
                            R, 
                            T,
                            compute_W=compute_W,
                            W=W,)
    if A.is_zero():
        IA.storage["P_red_to_zero"][A] = (W, IntegralPolynomial(0))
    return W, A

 

def reduce(IA:IntegralAlgebra, 
           Q:IntegralPolynomial, 
           T:OrderedSet[IntegralPolynomial],
           reduce_LM_only=True,
           compute_W=True) -> tuple[list,IntegralPolynomial]:
    """
    example: 
    * The list W = [(w0,P), (w1, w2,P)]
        encodes the integral polynomial
        w0*P + w1*int(w2*P)
    * Q = (w0*P + w1*int(w2*P)) + R
    * R corresponds to the polynomial Q reduced by the set T
    """
    if reduce_LM_only: 
        W,R = reduce_LM(IA, Q, T, compute_W=compute_W)
    else: 
        W,R = reduce_full(IA, Q, T, compute_W=compute_W) 
    return W, R



def __auto_reduce(IA:IntegralAlgebra, 
                T:OrderedSet[IntegralPolynomial],
                reduce_LM_only=True) -> tuple:
    T_reduced =  OrderedSet([])
    T_done = OrderedSet([]) 
    one_P_has_been_reduced = False
    for P in T:
        T_done.add(P) 
        T_copy = T - T_done | T_reduced 
        _, P_reduced = reduce(IA, P, T_copy, reduce_LM_only, compute_W=False)
        has_been_reduced = P_reduced != P  
        one_P_has_been_reduced = one_P_has_been_reduced or has_been_reduced 
        if not P_reduced.is_zero(): 
            T_reduced.add(P_reduced) 
        
    return T_reduced, one_P_has_been_reduced

def auto_reduce(IA:IntegralAlgebra, 
                T:OrderedSet[IntegralPolynomial],
                reduce_LM_only:bool=True,
                disable_deletion:bool=False
            ) -> OrderedSet[IntegralPolynomial]:
    T_reduced = T
    has_been_reduced = True 
    while has_been_reduced:  
        T_temp, has_been_reduced = __auto_reduce(IA,
                                                T_reduced,
                                                reduce_LM_only=reduce_LM_only) 
        T_reduced = T_temp  
    if disable_deletion:
        return T | T_reduced 
    else:
        return T_reduced
    






############################# 
# reduction utils functions #
#############################

def add_prefix_to_quotients(IA, prefix, W):
    """
    Example:
    The list W = [(w0,P), (w1, w2,P)]
        encodes the integral polynomial
        w0*P + w1*int(w2*P)
    This function add a prefix to the previous polynomial
    Take prefix = IM(x(t),y(t)**2)
    We compute
    x(t)*int(y(t)**2*int (w0*p)) + x(t)*int(y(t)**2*int(w1*int(w2*P))
    The following list is then returned:

    W_prime = [(IntPol(x(t)), IntPol(y(t)**2),w0,P), 
                (IntPol(x(t)),IntPol(y(t)**2),w1,w2,P)]
    """ 
    if W==[]:
        return W

    # convert prefix (which is an IM) into a list of integral polynomials
    # e.g : IM(x(t),y(t)) -> [IntegralPol(Im(x(t))),IntegralPol(IM(y(t))]
    prefix_pol_list = [
                IntegralPolynomial(IM(elem)) 
                for elem in prefix.get_content()
        ]
    W_prime = [] 
    
    for Wi in W:
        # Wi contains integral polynomials 
        W_prime += [(*prefix_pol_list, *Wi,)] 
    return W_prime



def mult_quotients_by_coeff(IA, W, coeff):  
    """
    Example:
    Consider a list 
    W = [(IntPol(IM(x(t))), IntPol(IM(y(t)**2)),w0,P), 
                (IntPol(IM(x(t))),IntPol(IM(y(t)**2)),w1,w2,P)]
    where w0,w1,w2 are Integral Polynomials.
    W encodes : x(t)*int(y(t)**2*int (w0*p)) + x(t)*int(y(t)**2*int(w1*int(w2*P))
    
    This function multiply the first polynomial of each tuple by coeff.

    The list 
        W_prime = [(IntPol(coeff*IM(x(t))), IntPol(IM(y(t)**2)),w0,P), 
                (IntPol(coeff*IM(x(t))),IntPol(IM(y(t)**2)),w1,w2,P)]
    is then returned.
    W_prime encodes :
        coeff*x(t)*int(y(t)**2*int (w0*p)) + coeff*x(t)*int(y(t)**2*int(w1*int(w2*P))
    """ 
    if W==[]:
        return W
    coeff_pol = IntegralPolynomial(coeff*IM(1))
    W_prime = []   
    for Wi in W: 
        W_prime += [(IA.polynomials_product(Wi[0],coeff_pol),
                    *Wi[1:],),] 
    return W_prime


import sympy as sp
import unittest
from IntegralElimination import *
from IntegralElimination.reduction import reduction_M_by_P_red_power_and_half_red_product

class TestIntegralElimination(unittest.TestCase):
    def test_15(self):
        x = sp.Function('x')
        y = sp.Function('y')
        t = sp.Symbol("t")
        theta = sp.Symbol("theta") 

        IA = IntegralAlgebra(order=[x(t),y(t)],
                            parameters=[theta])
        
        #is lm(P) is a monomial (not integral)
        P = IntegralPolynomial(IM(x(t))-IM(y(t)))
        M = IM(x(t)**2, y(t),x(t)*y(t))
        res = reduction_M_by_P_red_power_and_half_red_product(IA,M,P) 
        expected = None
        verify = expected == res
        self.assertTrue(verify) 

        # other test
        P = IntegralPolynomial(IM(1,x(t)*y(t))-IM(y(t)))   
        M = IM(x(t)**2, x(t)*y(t)**2)
        R = reduction_M_by_P_red_power_and_half_red_product(
                                    IA,M,P)[1].get_sympy_repr()
        expected = IM(x(t)**2*y(t)**2)/2
        verify = sp.simplify(expected - R) == 0 
        self.assertTrue(verify)

        # other test
        P = IntegralPolynomial(IM(1,x(t)*y(t),y(t))-IM(y(t)))   
        M = IM(x(t)**2, x(t)*y(t)**2)
        res = reduction_M_by_P_red_power_and_half_red_product(IA,M,P)
        expected = None
        verify = expected == res
        self.assertTrue(verify)

if __name__ == '__main__':
    unittest.main()


from setuptools import setup

setup(
    name="IntegralElimination",
    version="0.3.1",  
    packages=['IntegralElimination'],
    install_requires=[
        "ordered-set>=4.1.0",
        "sympy>=1.11.1", 
        "typing>=3.7.4.3",
        "IPython",
        "numpy"
    ],
    author="Louis ROUSSEL",
    author_email="louis.roussel@univ-lille.fr",
    description="Algorithm for performing integral elimination for nonlinear integral equations",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url="https://codeberg.org/louis-roussel/IntegralElimination", 
    # sympy needs python 3.8 but
    # the type annotations like tuple[IM, sp.Expr] requires
    # python 3.9
    python_requires='>=3.9', 
)
